/*
	G. Wiesenekker. ZZZZZZ a chess program.
	Copyright (C) 1993  G. Wiesenekker
	E-mail: wiesenecker@sara.nl

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _ZZZZZZ_H
#define _ZZZZZZ_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <math.h>

/*
Machine dependencies are treated by #if MACHINE == 
ZZZZZZ knows only about MAC and not MAC. That is, #define MACHINE DOS
or RS6000 or CRAY or SGI are all equivalent.
*/

#define MAC    0
#define DOS    1
#define RS6000 2
#define CRAY   3
#define SGI    4

#define MACHINE RS6000

/*
Setting DEBUG to 1 includes debugging code in the program.
Makes it run slow, but it is very useful if you run into problems.

Setting VERBOSE to 1 causes the static evaluation to print something
when it 'sees' something.

Setting OUTPUT to 1 causes the program to echo all it's output
to a file called 'echo' (useful on a MAC, less useful on UNIX
machines)
*/

#define DEBUG   0
#define VERBOSE 1
#define OUTPUT  1

/* 
NULL_MOVE_DEPTH is the minimum search depth at which we may
do a null move. We search regular moves at (depth - 1), 
but null moves are searched at (depth - 1 - NULL_MOVE_REDUC).
If the null move score fails high, the node is pruned.
(This fails terribly in case of zugzwang).
*/

#define NULL_MOVE_DEPTH 3
#define NULL_MOVE_REDUC 1

/*
THREAT_DEPTH is the minimum search depth at which we 
check for threats (see Ananthamaran, ICCA journal 19xx).
When the score fails high, we check if the null move score
is below (current material balance - THREAT_MARGIN).
If so, we extend the move that failed high. Threat extensions
may lead to infinite extensions (when white replies to a threat
with a threat, to which black replies with a threat etc.)
so we allow only THREAT_MAX extensions in a line of play.
As this makes the evaluation of a position history
dependent (it depends the number of threat extensions)
we include the number of threat extensions in the hash key.
*/

#define THREAT_DEPTH 3
#define THREAT_REDUC 1
#define THREAT_MARGIN (PAWN_SCORE * 2 + PAWN_SCORE / 2)
#define NTHREAT_MAX 2

/*
DepthMax is the maximum search depth.
TimeLimit is the allocated time.
If Append is TRUE we may append moves to the game file.
If UseBook is TRUE we may use the book.
GameTimeUsed is the total time used for the game
*/

extern int DepthMax;
extern int TimeLimit;
extern int Append;
extern int UseBook;
extern int GameTimeUsed;

/*
Some general macro's used by every C programmer.
*/

#define FALSE 0
#define TRUE 1
#define INVALID (-1)
#define LINE_MAX 256

/*
I have had some very nasty bugs in the past by typing & instead of &&.
*/

#define and &&
#define or  ||

#define ABS(A) ((A) > 0 ? (A) : -(A))
#define MAX(A, B) ((A) > (B) ? (A) : (B))
#define BIT(X)  (1 << (X))
#define BITL(X) (1L << (X))

#define FATAL_ERROR(X)\
	{my_printf("%s::%lu:%s\n", __FILE__, (ulong_t) __LINE__, X); exit(EXIT_FAILURE);}
	
/*
You will find many BUG macro's in the code. Of course, the error messages
are only understandable if you look at the code.
*/

#define BUG(X) {if (X) FATAL_ERROR(#X)}
#define MALLOC(P, T, N) BUG(((P) = (T *) malloc(sizeof(T) * (N))) == NULL)

typedef signed char byte_t;
typedef unsigned long ulong_t;

enum {WHITE = 0, BLACK, NONE, SIDE};
extern char *colour_as_string[2];
extern char *book_name[2];

/*
The value of the pieces, which are the normal chess values
1, 3, 3, 5, 10 scaled by a factor of 300 (to allow for some
flexibility in the evaluation function parameters).
*/

#define PAWN_SCORE    300
#define KNIGHT_SCORE  900
#define BISHOP_SCORE  900
#define ROOK_SCORE   1500
#define QUEEN_SCORE  3000
#define KING_SCORE      0

/*
The following macro's define bits for the different kind of moves
in chess.
*/

#define NULL_MOVE   BITL(16)
#define PROM_CAPT   BITL(17)
#define PROM_NORM   BITL(18)
#define CAPT        BITL(19)
#define EP          BITL(20)
#define SHORT       BITL(21)
#define LONG        BITL(22)
#define DOUBLE      BITL(23)
#define NORM        BITL(24)
#define PROM_KNIGHT BITL(25)
#define PROM_BISHOP BITL(26)
#define PROM_ROOK   BITL(27)
#define PROM_QUEEN  BITL(28)
#define INVAL_MOVE  BITL(29)

/*
In a future version I will squeeze the move in an int (6 bits from,
6 bits to, 4 bits kind).
*/

typedef long move_t;

/*
We use 64 bit hash keys. (See Marsland, ICCA 19xx).
*/

typedef struct
{
	ulong_t lo, hi;
} hash_key_t;

/*
We store the pieces in lists, one list for every piece, so that
we can first generate the pawn moves, then the knight moves etc.
(Note: I usually prefix the names of the fields in a struct with
the name of the struct).
board_colour is WHITE, BLACK, SIDE or NONE.
board_kind is INVALID, PAWN, KNIGHT, ROOK, QUEEN or KING.
If board_kind is INVALID, the end of the list is reached.
board_value is the material value of the piece.
board_where tells where the piece is on the board.
if board_where is INVALID, the piece has been captured.
So you will see often this loop:

board_t *p;
for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
{
	if (p->board_where == INVALID) continue;
	..
}
board_moved tells whether the piece has moved (useful for
checking castling rights and pieces that have not been
developed).
board_captured tells whether the piece has captured (useful
in the quiescence search).
board_key is a pointer to a table with hash keys.
board_dir is a pointer to a table of offsets for the 
different directions in which a piece can move.
board_shift tells wheter we may offset the piece more than
once (FALSE for PAWN, KNIGHT and KING, TRUE for BISHOP,
ROOK and QUEEN).
board_count is a pointer to a table with piece counts.
board_strength is the strength value of the piece 
(see ICCA july 1993).
*/

typedef struct
{
	int board_colour, board_kind, board_value, board_where;
	int board_moved, board_captured;
	hash_key_t *board_key;
	int board_shift, *board_dir;
	int *board_count;
	int board_strength;
} board_t;

/*
A node contains all kinds of information about a node.
When generating the captures, we also generate the non-capturing
moves. move_base is a pointer to a table where we store the non-capturing
moves, capt_base is a pointer to a table where we store the capturing
moves.
We need capt and prom for undoing captures and promotions.
We need node_moved and node_captured for resetting the
moved and captured attributes when undoing a move.
node_reverse tells whether the move was reversible (useful for
checking for draw by repetition).
node_hack_ep contains the square where we may take en passant (will
be INVALID most of the time).
node_move is the move itself.
*/

typedef struct
{
	move_t *move_base, *capt_base;
	board_t *capt, *prom;
	int node_moved, node_captured;
	hash_key_t node_key, node_total_key;
	int node_mat[2], node_pos[2];
	int node_reverse;
	int node_hack_ep;
	move_t node_move;
} node_t;

/*
We store the hash keys in the opening book. As there are only
2^64 keys, and many more chess positions, we also store the number
of pieces in the book to guard against hash faults.
book_hit is the number of times the position occurs in the opening book.
*/

typedef struct
{
	int book_tot_count[2];
	hash_key_t book_total_key;
	int book_hit;
} book_t;

extern int tot_count[2];
extern node_t *node;
extern book_t *book[2];
extern long nbook[2];
extern char *book_name[2];

/*
The legal moves in the starting position are stored in array pv.
pv_move is the move.
pv_string is the move in string form.
pv_score is the move score (not very useful, because it are lower
and upper bounds for most moves).
pv_hit is the number of times the move occurs in an opening book
variation.
*/

#define STRING_MAX 16

typedef struct
{
	move_t pv_move;
	char pv_string[STRING_MAX];
	int pv_score;
	int pv_hit;
} pv_t;

#define NPVS_MAX 256

extern int npv;
extern pv_t pv[NPVS_MAX];
extern int jpv;
extern int GameTimeUsed;

/* 
prototypes
*/

void my_printf(char *, ...);
ulong_t ulrand(void);
int read_board(FILE *);
void display_board(void);
void write_board(FILE *f);
void display_score(void);
void display_list(void);
void display_occup(void);
void display_attack(void);
void compact_lists(void);
void update_key(int, hash_key_t *);
int attacked(int, int);
int i_am_check(int, int);
void do_move(int, int, move_t);
void undo_move(int, int);
void set_score(double *);
int draw_by_repetition(int);
void gen_pv(int, int);
int search_pv(char *, int);
int get_book_move(int, int);
void do_pv(int, int);
void sort_pv(int, int);

void read_game(FILE *, int *, int *);
void analyze_pos(char *);
void analyze_game(char *);
double solve_problems(char *);
void play_batch(char *);
void play_game(int, int, char *);
void do_what(char *, char *);
void play_par(char *, char *);
void gen_book(char *);
void optimise(char *);
void init(void);

#endif
